@@ -56,9 +56,9 @@ import java.util.HashMap;
 import java.util.HashSet;
 import java.util.List;
 import java.util.Map;
-import java.util.UUID;
 import java.util.Map.Entry;
 import java.util.Set;
+import java.util.UUID;
 import java.util.concurrent.ConcurrentHashMap;
 import java.util.concurrent.ExecutionException;
 import java.util.concurrent.ExecutorService;
@@ -71,11 +71,11 @@ import com.hazelcast.core.HazelcastInstance;
 import com.hazelcast.core.Member;
 import com.openexchange.exception.OXException;
 import com.openexchange.java.util.UUIDs;
-import com.openexchange.realtime.RealtimeExceptionCodes;
 import com.openexchange.realtime.directory.Resource;
 import com.openexchange.realtime.directory.ResourceDirectory;
 import com.openexchange.realtime.dispatch.LocalMessageDispatcher;
 import com.openexchange.realtime.dispatch.MessageDispatcher;
+import com.openexchange.realtime.exception.RealtimeExceptionCodes;
 import com.openexchange.realtime.hazelcast.Services;
 import com.openexchange.realtime.hazelcast.channel.HazelcastAccess;
 import com.openexchange.realtime.hazelcast.channel.StanzaDispatcher;
@@ -107,6 +107,7 @@ public class GlobalMessageDispatcherImpl implements MessageDispatcher {
     @Override
     public Stanza sendSynchronously(Stanza stanza, long timeout, TimeUnit unit) throws OXException {
         String uuid = UUIDs.getUnformattedString(UUID.randomUUID());
+        stanza.trace("Send synchronously. UUID: " + uuid);
         channel.setUp(uuid, stanza);
         send(stanza);
         return channel.waitFor(uuid, timeout, unit);
@@ -128,6 +129,7 @@ public class GlobalMessageDispatcherImpl implements MessageDispatcher {
         }
 
         if (recipients == null) {
+            stanza.trace("Parameter 'recipients' must not be null!");
             throw new IllegalArgumentException("Parameter 'recipients' must not be null!");
         }
 
@@ -163,6 +165,7 @@ public class GlobalMessageDispatcherImpl implements MessageDispatcher {
         if (localIds != null) {
             // Send via local message dispatcher to locally reachable receivers
             ensureSequence(stanza, localMember);
+            stanza.trace("Deliver locally");
             Map<ID, OXException> sent = Services.getService(LocalMessageDispatcher.class).send(stanza, localIds);
             exceptions.putAll(sent);
         }
@@ -172,6 +175,7 @@ public class GlobalMessageDispatcherImpl implements MessageDispatcher {
         for (Member receiver : targets.keySet()) {
             Set<ID> ids = targets.get(receiver);
             LOG.debug("Sending to '" + stanza.getTo() + "' @ " + receiver);
+            stanza.trace("Sending to '" + stanza.getTo() + "' @ " + receiver);
             ensureSequence(stanza, receiver);
             FutureTask<Map<ID, OXException>> task = new DistributedTask<Map<ID, OXException>>(new StanzaDispatcher(stanza, ids) , receiver);
             executorService.execute(task);
@@ -189,7 +193,9 @@ public class GlobalMessageDispatcherImpl implements MessageDispatcher {
                 throw ThreadPools.launderThrowable(e, OXException.class);
             }
         }
-
+        if (!exceptions.isEmpty()) {
+            stanza.trace(exceptions);
+        }
         return exceptions;
     }
     
@@ -221,6 +227,7 @@ public class GlobalMessageDispatcherImpl implements MessageDispatcher {
                 nextNumber = (otherNextNumber != null) ? otherNextNumber : nextNumber;
             }
             stanza.setSequenceNumber(nextNumber.incrementAndGet() - 1);
+            stanza.trace("Updating sequence number for " + receiver + ": " + stanza.getSequenceNumber());
         }
     }
 }
